# relative
from .deps import DependencyGridGit
from .deps import check_deps


def verify_git_installation() -> None:
    dep = DependencyGridGit(name="git", output_in_text=True)
    deps = {}
    deps["git"] = dep
    check_deps(of="Git", deps=deps, display=False, output_in_text=True)  # type: ignore
    if dep.issues:
        exit(1)


verify_git_installation()



"""Python Level API to launch Docker Containers using Hagrid"""
# future
from __future__ import annotations

# stdlib
from enum import Enum
import getpass
import inspect
import os
import subprocess  # nosec
from typing import Any
from typing import Callable
from typing import Optional
from typing import Union

# third party
import gevent

# relative
from .cli import str_to_bool
from .grammar import find_available_port
from .names import random_name
from .util import shell

try:
    # syft absolute
    from syft.abstract_node import NodeSideType
    from syft.abstract_node import NodeType
    from syft.node.node import get_default_root_password
    from syft.util.util import prompt_warning_message
except Exception:  # nosec
    # print("Please install syft with `pip install syft`")
    pass

DEFAULT_PORT = 8080
# Gevent used instead of threading module ,as we monkey patch gevent in syft
# and this causes context switch error when we use normal threading in hagrid


# Define a function to read and print a stream
def read_stream(stream: subprocess.PIPE) -> None:
    while True:
        line = stream.readline()
        if not line:
            break
        print(line, end="")
        gevent.sleep(0)


def to_snake_case(name: str) -> str:
    return name.lower().replace(" ", "_")


def get_syft_client() -> Optional[Any]:
    try:
        # syft absolute
        import syft as sy

        return sy
    except Exception:  # nosec
        # print("Please install syft with `pip install syft`")
        pass
    return None


def container_exists(name: str) -> bool:
    output = shell(f"docker ps -q -f name='{name}'")
    return len(output) > 0


def container_id(name: str) -> Optional[str]:
    output = shell(f"docker ps -q -f name='{name}'")
    if len(output) > 0:
        return output[0].strip()
    return None


def port_from_container(name: str) -> Optional[int]:
    cid = container_id(name)
    if cid is None:
        return None
    output = shell(f"docker port {cid}")
    if len(output) > 0:
        try:
            # 80/tcp -> 0.0.0.0:8080
            lines = output.split("\n")
            parts = lines[0].split(":")
            port = int(parts[1].strip())
            return port
        except Exception:  # nosec
            return None
    return None


def container_exists_with(name: str, port: int) -> bool:
    output = shell(f"docker ps -q -f name='{name}' -f expose='{port}'")
    return len(output) > 0


def get_node_type(node_type: Optional[Union[str, NodeType]]) -> Optional[NodeType]:
    if node_type is None:
        node_type = os.environ.get("ORCHESTRA_NODE_TYPE", NodeType.DOMAIN)
    try:
        return NodeType(node_type)
    except ValueError:
        print(f"node_type: {node_type} is not a valid NodeType: {NodeType}")
    return None


def get_deployment_type(deployment_type: Optional[str]) -> Optional[DeploymentType]:
    if deployment_type is None:
        deployment_type = os.environ.get(
            "ORCHESTRA_DEPLOYMENT_TYPE", DeploymentType.PYTHON
        )

    # provide shorthands
    if deployment_type == "container":
        deployment_type = "container_stack"

    try:
        return DeploymentType(deployment_type)
    except ValueError:
        print(
            f"deployment_type: {deployment_type} is not a valid DeploymentType: {DeploymentType}"
        )
    return None


# Can also be specified by the environment variable
# ORCHESTRA_DEPLOYMENT_TYPE
class DeploymentType(Enum):
    PYTHON = "python"
    SINGLE_CONTAINER = "single_container"
    CONTAINER_STACK = "container_stack"
    K8S = "k8s"


class NodeHandle:
    def __init__(
        self,
        node_type: NodeType,
        deployment_type: DeploymentType,
        node_side_type: NodeSideType,
        name: str,
        port: Optional[int] = None,
        url: Optional[str] = None,
        python_node: Optional[Any] = None,
        shutdown: Optional[Callable] = None,
    ) -> None:
        self.node_type = node_type
        self.name = name
        self.port = port
        self.url = url
        self.python_node = python_node
        self.shutdown = shutdown
        self.deployment_type = deployment_type
        self.node_side_type = node_side_type

    @property
    def client(self) -> Any:
        if self.port:
            sy = get_syft_client()
            return sy.login(url=self.url, port=self.port, verbose=False)  # type: ignore
        elif self.deployment_type == DeploymentType.PYTHON:
            return self.python_node.get_guest_client(verbose=False)  # type: ignore
        else:
            raise NotImplementedError(
                f"client not implemented for the deployment type:{self.deployment_type}"
            )

    def login(
        self, email: Optional[str] = None, password: Optional[str] = None, **kwargs: Any
    ) -> Optional[Any]:
        if password == get_default_root_password():
            message = (
                "You are using a default password. Please change the password "
                "using `[your_client].me.set_password([new_password])`."
            )
            prompt_warning_message(message)
        client = self.client
        if email and password:
            return client.login(email=email, password=password, **kwargs)
        return None

    def register(
        self,
        name: str,
        email: str,
        password: str,
        institution: Optional[str] = None,
        website: Optional[str] = None,
    ) -> Any:
        client = self.client
        return client.register(
            name=name,
            email=email,
            password=password,
            institution=institution,
            website=website,
        )

    def land(self) -> None:
        if self.deployment_type == DeploymentType.PYTHON:
            if self.shutdown:
                self.shutdown()
        else:
            Orchestra.land(self.name, deployment_type=self.deployment_type)


def deploy_to_python(
    node_type_enum: NodeType,
    deployment_type_enum: DeploymentType,
    port: Union[int, str],
    name: str,
    host: str,
    reset: bool,
    tail: bool,
    dev_mode: bool,
    processes: int,
    local_db: bool,
    node_side_type: NodeSideType,
    enable_warnings: bool,
) -> Optional[NodeHandle]:
    sy = get_syft_client()
    if sy is None:
        return sy
    worker_classes = {NodeType.DOMAIN: sy.Domain, NodeType.NETWORK: sy.Gateway}

    # syft >= 0.8.2
    if hasattr(sy, "Enclave"):
        worker_classes[NodeType.ENCLAVE] = sy.Enclave
    if hasattr(NodeType, "GATEWAY"):
        worker_classes[NodeType.GATEWAY] = sy.Gateway

    if port:
        if port == "auto":
            # dont use default port to prevent port clashes in CI
            port = find_available_port(host="localhost", port=None, search=True)
        sig = inspect.signature(sy.serve_node)
        if "node_type" in sig.parameters.keys():
            start, stop = sy.serve_node(
                name=name,
                host=host,
                port=port,
                reset=reset,
                dev_mode=dev_mode,
                tail=tail,
                node_type=node_type_enum,
                node_side_type=node_side_type,
                enable_warnings=enable_warnings,
            )
        else:
            # syft <= 0.8.1
            start, stop = sy.serve_node(
                name=name,
                host=host,
                port=port,
                reset=reset,
                dev_mode=dev_mode,
                tail=tail,
            )
        start()
        return NodeHandle(
            node_type=node_type_enum,
            deployment_type=deployment_type_enum,
            name=name,
            port=port,
            url="http://localhost",
            shutdown=stop,
            node_side_type=node_side_type,
        )
    else:
        if node_type_enum in worker_classes:
            worker_class = worker_classes[node_type_enum]
            sig = inspect.signature(worker_class.named)
            if "node_type" in sig.parameters.keys():
                worker = worker_class.named(
                    name=name,
                    processes=processes,
                    reset=reset,
                    local_db=local_db,
                    node_type=node_type_enum,
                    node_side_type=node_side_type,
                    enable_warnings=enable_warnings,
                )
            else:
                # syft <= 0.8.1
                worker = worker_class.named(
                    name=name,
                    processes=processes,
                    reset=reset,
                    local_db=local_db,
                )
        else:
            raise NotImplementedError(f"node_type: {node_type_enum} is not supported")
        return NodeHandle(
            node_type=node_type_enum,
            deployment_type=deployment_type_enum,
            name=name,
            python_node=worker,
            node_side_type=node_side_type,
        )


def deploy_to_k8s(
    node_type_enum: NodeType,
    deployment_type_enum: DeploymentType,
    name: str,
    node_side_type: NodeSideType,
) -> NodeHandle:
    node_port = int(os.environ.get("NODE_PORT", f"{DEFAULT_PORT}"))
    return NodeHandle(
        node_type=node_type_enum,
        deployment_type=deployment_type_enum,
        name=name,
        port=node_port,
        url="http://localhost",
        node_side_type=node_side_type,
    )


def deploy_to_container(
    node_type_enum: NodeType,
    deployment_type_enum: DeploymentType,
    node_side_type: NodeSideType,
    reset: bool,
    cmd: bool,
    tail: bool,
    verbose: bool,
    tag: str,
    render: bool,
    dev_mode: bool,
    port: Union[int, str],
    name: str,
    enable_warnings: bool,
) -> Optional[NodeHandle]:
    if port == "auto" or port is None:
        if container_exists(name=name):
            port = port_from_container(name=name)  # type: ignore
        else:
            port = find_available_port(host="localhost", port=DEFAULT_PORT, search=True)

    # Currently by default we launch in dev mode
    if reset:
        Orchestra.reset(name, deployment_type_enum)
    else:
        if container_exists_with(name=name, port=port):
            return NodeHandle(
                node_type=node_type_enum,
                deployment_type=deployment_type_enum,
                name=name,
                port=port,
                url="http://localhost",
                node_side_type=node_side_type,
            )

    # Start a subprocess and capture its output
    commands = ["hagrid", "launch"]

    name = random_name() if not name else name
    commands.extend([name, node_type_enum.value])

    commands.append("to")
    commands.append(f"docker:{port}")

    if dev_mode:
        commands.append("--dev")

    if not enable_warnings:
        commands.append("--no-warnings")

    # by default , we deploy as container stack
    if deployment_type_enum == DeploymentType.SINGLE_CONTAINER:
        commands.append("--deployment-type=single_container")

    if cmd:
        commands.append("--cmd")

    if tail:
        commands.append("--tail")

    if verbose:
        commands.append("--verbose")

    if tag:
        commands.append(f"--tag={tag}")

    if render:
        commands.append("--render")

    # needed for building containers
    USER = os.environ.get("USER", getpass.getuser())
    env = os.environ.copy()
    env["USER"] = USER

    process = subprocess.Popen(  # nosec
        commands, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, env=env
    )
    # Start gevent threads to read and print the output and error streams
    stdout_thread = gevent.spawn(read_stream, process.stdout)
    stderr_thread = gevent.spawn(read_stream, process.stderr)

    # Wait for the threads to finish
    gevent.joinall([stdout_thread, stderr_thread], raise_error=True)

    if not cmd:
        return NodeHandle(
            node_type=node_type_enum,
            deployment_type=deployment_type_enum,
            name=name,
            port=port,
            url="http://localhost",
            node_side_type=node_side_type,
        )
    return None


class Orchestra:
    @staticmethod
    def launch(
        # node information and deployment
        name: Optional[str] = None,
        node_type: Optional[Union[str, NodeType]] = None,
        deploy_to: Optional[str] = None,
        node_side_type: Optional[str] = None,
        # worker related inputs
        port: Optional[Union[int, str]] = None,
        processes: int = 1,  # temporary work around for jax in subprocess
        local_db: bool = False,
        dev_mode: bool = False,
        cmd: bool = False,
        reset: bool = False,
        tail: bool = False,
        host: Optional[str] = "0.0.0.0",  # nosec
        tag: Optional[str] = "latest",
        verbose: bool = False,
        render: bool = False,
        enable_warnings: bool = False,
    ) -> Optional[NodeHandle]:
        if dev_mode is True:
            os.environ["DEV_MODE"] = "True"

        # syft 0.8.1
        if node_type == "python":
            node_type = NodeType.DOMAIN
            if deploy_to is None:
                deploy_to = "python"

        dev_mode = str_to_bool(os.environ.get("DEV_MODE", f"{dev_mode}"))

        node_type_enum: Optional[NodeType] = get_node_type(node_type=node_type)
        if not node_type_enum:
            return None

        node_side_type_enum = (
            NodeSideType.HIGH_SIDE
            if node_side_type is None
            else NodeSideType(node_side_type)
        )

        deployment_type_enum: Optional[DeploymentType] = get_deployment_type(
            deployment_type=deploy_to
        )
        if not deployment_type_enum:
            return None

        if deployment_type_enum == DeploymentType.PYTHON:
            return deploy_to_python(
                node_type_enum=node_type_enum,
                deployment_type_enum=deployment_type_enum,
                port=port,
                name=name,
                host=host,
                reset=reset,
                tail=tail,
                dev_mode=dev_mode,
                processes=processes,
                local_db=local_db,
                node_side_type=node_side_type_enum,
                enable_warnings=enable_warnings,
            )

        elif deployment_type_enum == DeploymentType.K8S:
            return deploy_to_k8s(
                node_type_enum=node_type_enum,
                deployment_type_enum=deployment_type_enum,
                name=name,
                node_side_type=node_side_type_enum,
            )

        elif (
            deployment_type_enum == DeploymentType.CONTAINER_STACK
            or deployment_type_enum == DeploymentType.SINGLE_CONTAINER
        ):
            return deploy_to_container(
                node_type_enum=node_type_enum,
                deployment_type_enum=deployment_type_enum,
                reset=reset,
                cmd=cmd,
                tail=tail,
                verbose=verbose,
                tag=tag,
                render=render,
                dev_mode=dev_mode,
                port=port,
                name=name,
                node_side_type=node_side_type_enum,
                enable_warnings=enable_warnings,
            )
        else:
            print(f"deployment_type: {deployment_type_enum} is not supported")
            return None

    @staticmethod
    def land(
        name: str, deployment_type: Union[str, DeploymentType], reset: bool = False
    ) -> None:
        deployment_type_enum = DeploymentType(deployment_type)
        Orchestra.shutdown(name=name, deployment_type_enum=deployment_type_enum)
        if reset:
            Orchestra.reset(name, deployment_type_enum=deployment_type_enum)

    @staticmethod
    def shutdown(name: str, deployment_type_enum: DeploymentType) -> None:
        if deployment_type_enum != DeploymentType.PYTHON:
            snake_name = to_snake_case(name)

            land_output = shell(f"hagrid land {snake_name} --force")
            if "Removed" in land_output:
                print(f" ✅ {snake_name} Container Removed")
            else:
                print(f"❌ Unable to remove container: {snake_name} :{land_output}")

    @staticmethod
    def reset(name: str, deployment_type_enum: DeploymentType) -> None:
        if deployment_type_enum == DeploymentType.PYTHON:
            sy = get_syft_client()
            _ = sy.Worker.named(name, processes=1, reset=True)  # type: ignore
        elif (
            deployment_type_enum == DeploymentType.CONTAINER_STACK
            or deployment_type_enum == DeploymentType.SINGLE_CONTAINER
        ):
            if container_exists(name=name):
                Orchestra.shutdown(name=name, deployment_type_enum=deployment_type_enum)

            snake_name = to_snake_case(name)

            volume_output = shell(
                f"docker volume rm {snake_name}_credentials-data --force || true"
            )

            if "Error" not in volume_output:
                print(f" ✅ {snake_name} Volume Removed")
            else:
                print(
                    f"❌ Unable to remove container volume: {snake_name} :{volume_output}"
                )
        else:
            raise NotImplementedError(
                f"Reset not implemented for the deployment type:{deployment_type_enum}"
            )



# stdlib
import hashlib
import os
import shutil
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from urllib.parse import urlparse

# third party
from jinja2 import Environment
from jinja2 import FileSystemLoader
from jinja2 import Template
import requests
from rich.progress import track
import yaml

# relative
from .cache import DEFAULT_REPO
from .cache import STABLE_BRANCH
from .lib import hagrid_cache_dir
from .lib import manifest_template_path
from .lib import repo_src_path
from .mode import EDITABLE_MODE

HAGRID_TEMPLATE_PATH = str(manifest_template_path())


def read_yml_file(filename: str) -> Tuple[Optional[Dict], str]:
    template = None

    with open(filename) as fp:
        try:
            text = fp.read()
            template = yaml.safe_load(text)
            template_hash = hashlib.sha256(text.encode("utf-8")).hexdigest()
        except yaml.YAMLError as exc:
            raise exc

    return template, template_hash


def read_yml_url(yml_url: str) -> Tuple[Optional[Dict], str]:
    template = None

    try:
        # download file
        response = requests.get(yml_url)  # nosec
        if response.status_code != 200:
            raise Exception(f"Failed to download: {yml_url}")

        # Save file to the local destination
        try:
            template = yaml.safe_load(response.content)
            template_hash = hashlib.sha256(response.content).hexdigest()
        except yaml.YAMLError as exc:
            raise exc

    except Exception as e:
        raise e

    return template, template_hash


def git_url_for_file(file_path: str, base_url: str, hash: str) -> str:
    # url must have unix style slashes
    return os.path.join(base_url, hash, file_path).replace(os.sep, "/")


def get_local_abs_path(target_dir: str, file_path: str) -> str:
    local_path = os.path.join(target_dir, file_path)
    return os.path.expanduser(local_path)


def is_url(string: str) -> bool:
    try:
        result = urlparse(string)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def is_path(string: str) -> bool:
    return os.path.exists(string)


def manifest_cache_path(template_hash: str) -> str:
    return f"{hagrid_cache_dir()}/manifests/{template_hash}"


def url_from_repo(template_location: Optional[str]) -> Optional[str]:
    if template_location is None:
        return None

    if ":" in template_location and "/" in template_location:
        parts = template_location.split(":")
        branch_or_hash = parts[1]
        repo = parts[0]
    elif ":" not in template_location and "/" in template_location:
        branch_or_hash = STABLE_BRANCH
        repo = template_location
    else:
        branch_or_hash = template_location
        repo = DEFAULT_REPO

    manifest_url = (
        f"https://raw.githubusercontent.com/{repo}/{branch_or_hash}"
        "/packages/hagrid/hagrid/manifest_template.yml"
    )

    if is_url(manifest_url):
        return manifest_url
    return None


def get_template_yml(template_location: Optional[str]) -> Tuple[Optional[Dict], str]:
    if template_location:
        if is_url(template_location):
            template, template_hash = read_yml_url(template_location)
        elif is_path(template_location):
            template, template_hash = read_yml_file(template_location)
        elif url_from_repo(template_location):
            template, template_hash = read_yml_url(url_from_repo(template_location))
        else:
            raise Exception(f"{template_location} is not valid")
    else:
        template_location = HAGRID_TEMPLATE_PATH

        template, template_hash = read_yml_file(template_location)

    if EDITABLE_MODE and is_path(template_location):
        # save it to the same folder for dev mode
        template_hash = "dev"
    return template, template_hash


def setup_from_manifest_template(
    host_type: str,
    deployment_type: str,
    template_location: Optional[str] = None,
    overwrite: bool = False,
    verbose: bool = False,
) -> Dict:
    template, template_hash = get_template_yml(template_location)

    kwargs_to_parse = {}

    if template is None:
        raise ValueError(
            f"Failed to read {template_location}. Please check the file name or path is correct."
        )

    git_hash = template["hash"]
    git_base_url = template["baseUrl"]
    target_dir = manifest_cache_path(template_hash)
    all_template_files = template["files"]
    docker_tag = template["dockerTag"]
    files_to_download = []

    for package_name in all_template_files:
        # Get all files w.r.t that package e.g. grid, syft, hagrid
        template_files = all_template_files[package_name]
        package_path = template_files["path"]

        # common files
        files_to_download += [
            os.path.join(package_path, f) for f in template_files["common"]
        ]

        # worker
        if deployment_type == "single_container" and host_type in ["docker"]:
            files_to_download += [
                os.path.join(package_path, f) for f in template_files["worker"]
            ]
        # docker related files
        elif host_type in ["docker"]:
            files_to_download += [
                os.path.join(package_path, f) for f in template_files["docker"]
            ]

        # add k8s related files
        # elif host_type in ["k8s"]:
        #     files_to_download += template_files["k8s"]

        else:
            raise Exception(f"Hagrid template does not currently support {host_type}.")

    if EDITABLE_MODE and is_path(template_location):
        # to test things in editable mode we can pass in a .yml file path and it will
        # copy the files instead of download them
        for src_file_path in track(files_to_download, description="Copying files"):
            full_src_dir = f"{repo_src_path()}/{src_file_path}"
            full_target_path = f"{target_dir}/{src_file_path}"
            full_target_dir = os.path.dirname(full_target_path)
            os.makedirs(full_target_dir, exist_ok=True)

            shutil.copyfile(
                full_src_dir,
                full_target_path,
            )
    else:
        download_files(
            files_to_download=files_to_download,
            git_hash=git_hash,
            git_base_url=git_base_url,
            target_dir=target_dir,
            overwrite=overwrite,
            verbose=verbose,
        )

    kwargs_to_parse["tag"] = docker_tag
    return kwargs_to_parse


def deployment_dir(node_name: str) -> str:
    return f"{hagrid_cache_dir()}/deployments/{node_name}"


def download_files(
    files_to_download: List[str],
    git_hash: str,
    git_base_url: str,
    target_dir: str,
    overwrite: bool = False,
    verbose: bool = False,
) -> None:
    for src_file_path in track(files_to_download, description="Downloading files"):
        # For now target file path is same as source file path
        trg_file_path = src_file_path
        local_destination = get_local_abs_path(target_dir, trg_file_path)
        link_to_file = git_url_for_file(src_file_path, git_base_url, git_hash)
        download_file(
            link_to_file=link_to_file,
            local_destination=local_destination,
            overwrite=overwrite,
            verbose=verbose,
        )


def render_templates(
    node_name: str,
    deployment_type: str,
    template_location: Optional[str],
    env_vars: dict,
    host_type: str,
) -> None:
    template, template_hash = get_template_yml(template_location)

    if template is None:
        raise ValueError("Failed to read hagrid template.")

    src_dir = manifest_cache_path(template_hash)
    target_dir = deployment_dir(node_name)
    all_template_files = template["files"]

    jinja_template = JinjaTemplate(src_dir)

    files_to_render = []
    for package_name in all_template_files:
        template_files = all_template_files[package_name]

        # Aggregate all the files to be rendered

        # common files
        files_to_render += template_files["common"]

        # worker
        if deployment_type == "single_container" and host_type in ["docker"]:
            for template_file in template_files["worker"]:
                if "default.env" not in template_file:
                    files_to_render.append(template_file)

        elif host_type in ["docker"]:
            # docker related files
            for template_file in template_files["docker"]:
                if "default.env" not in template_file:
                    files_to_render.append(template_file)

        # Render the files
        for file_path in files_to_render:
            folder_path = template_files["path"]
            # relative to src_dir
            src_file_path = f"{folder_path}{file_path}"
            target_file_path = f"{target_dir}/{file_path}"
            os.makedirs(os.path.dirname(target_file_path), exist_ok=True)
            jinja_template.substitute_vars(src_file_path, env_vars, target_file_path)


class JinjaTemplate(object):
    def __init__(self, template_dir: Union[str, os.PathLike]) -> None:
        self.directory = os.path.expanduser(template_dir)
        self.environ = Environment(
            loader=FileSystemLoader(self.directory), autoescape=True
        )

    def read_template_from_path(self, filepath: str) -> Template:
        return self.environ.get_template(name=filepath)

    def substitute_vars(
        self, template_path: str, vars_to_substitute: dict, target_path: str
    ) -> None:
        template = self.read_template_from_path(template_path)
        rendered_template = template.render(vars_to_substitute)
        self.save_to(rendered_template, target_path)

    def save_to(self, message: str, filename: str) -> None:
        base_dir = self.directory
        filepath = os.path.abspath(os.path.join(base_dir, filename))

        # Create sub directories if does not exist
        os.makedirs(os.path.dirname(filepath), exist_ok=True)

        # Save template to filepath
        with open(filepath, "w") as fp:
            fp.write(message)


def download_file(
    link_to_file: str,
    local_destination: str,
    overwrite: bool = False,
    verbose: bool = False,
) -> None:
    file_dir = os.path.dirname(local_destination)
    os.makedirs(file_dir, exist_ok=True)

    if not os.path.exists(local_destination) or overwrite:
        try:
            # download file
            response = requests.get(link_to_file)  # nosec
            if response.status_code != 200:
                raise Exception(f"Failed to download: {link_to_file}")

            # Save file to the local destination
            open(local_destination, "wb").write(response.content)

        except Exception as e:
            raise e
    else:
        if verbose:
            print(f"Skipping download: {link_to_file} exists.")



# stdlib
import os
import subprocess  # nosec
import sys
from typing import Any
from typing import Tuple
from typing import Union
from urllib.parse import urlparse


def from_url(url: str) -> Tuple[str, str, int, str, Union[Any, str]]:
    try:
        # urlparse doesnt handle no protocol properly
        if "://" not in url:
            url = "http://" + url
        parts = urlparse(url)
        host_or_ip_parts = parts.netloc.split(":")
        # netloc is host:port
        port = 80
        if len(host_or_ip_parts) > 1:
            port = int(host_or_ip_parts[1])
        host_or_ip = host_or_ip_parts[0]
        return (
            host_or_ip,
            parts.path,
            port,
            parts.scheme,
            getattr(parts, "query", ""),
        )
    except Exception as e:
        print(f"Failed to convert url: {url} to GridURL. {e}")
        raise e


def fix_windows_virtualenv_api(cls: type) -> None:
    # fix bug in windows
    def _python_rpath(self: Any) -> str:
        """The relative path (from environment root) to python."""
        # Windows virtualenv installation installs pip to the [Ss]cripts
        # folder. Here's a simple check to support:
        if sys.platform == "win32":
            # fix here https://github.com/sjkingo/virtualenv-api/issues/47
            return os.path.join(self.path, "Scripts", "python.exe")
        return os.path.join("bin", "python")

    cls._python_rpath = property(_python_rpath)


def shell(command: str) -> str:
    try:
        output = subprocess.check_output(  # nosec
            command, shell=True, stderr=subprocess.STDOUT
        )
    except Exception:
        output = b""
    return output.decode("utf-8")



# stdlib
import os
from pathlib import Path
import site
from typing import Optional


def str_to_bool(bool_str: Optional[str]) -> bool:
    result = False
    bool_str = str(bool_str).lower()
    if bool_str == "true" or bool_str == "1":
        result = True
    return result


def hagrid_root() -> str:
    return os.path.abspath(str(Path(__file__).parent.parent))


def is_editable_mode() -> bool:
    disable_editable_mode = str_to_bool(
        os.environ.get("DISABLE_EDITABLE_MODE", "False")
    )
    if disable_editable_mode:
        print("🚨 Editable Mode DISABLED")
        return False
    current_package_root = hagrid_root()

    installed_as_editable = False
    sitepackages_dirs = site.getsitepackages()
    # check all site-packages returned if they have a hagrid.egg-link
    for sitepackages_dir in sitepackages_dirs:
        egg_link_file = Path(sitepackages_dir) / "hagrid.egg-link"
        try:
            linked_folder = egg_link_file.read_text()
            # if the current code is in the same path as the egg-link its -e mode
            installed_as_editable = current_package_root in linked_folder
            break
        except Exception:  # nosec
            pass

    if os.path.exists(Path(current_package_root) / "hagrid.egg-info"):
        installed_as_editable = True

    return installed_as_editable


EDITABLE_MODE = is_editable_mode()
